# -*- coding: utf-8 -*-
"""Decrypt objects in Config."""
from __future__ import unicode_literals
from future.utils import bytes_to_native_str as n

from base64 import b64decode
import logging
import os

import boto3
from botocore.exceptions import ClientError, NoRegionError

from figgypy.exceptions import FiggypyError
from figgypy.utils import env_or_default

log = logging.getLogger('figgypy')

GPG_IMPORTED = False
try:
    import gnupg
    GPG_IMPORTED = True
except ImportError:
    logging.info('Could not load gnupg. Will be unable to unpack secrets.')


def gpg_decrypt(cfg, gpg_config=None):
    """Decrypt GPG objects in configuration.

    Args:
        cfg (dict): configuration dictionary
        gpg_config (dict): gpg configuration
            dict of arguments for gpg including:
                homedir, binary, and keyring (require all if any)
            example:
                gpg_config = {'homedir': '~/.gnupg/',
                              'binary': 'gpg',
                              'keyring': 'pubring.kbx'}

    Returns:
        dict: decrypted configuration dictionary

    The aim is to find in the dictionary items which have been encrypted
    with gpg, then decrypt them if possible.

    We will either detect the encryption based on the PGP block text or a
    user can create a key "_gpg" in which to store the data. Either case
    will work. In the case of the "_gpg" key all data at this level will
    be replaced with the decrypted contents. For example:

        {'component': {'key': 'PGP Block ...'}}

    will transform to:

        {'component': {'key': 'decrypted value'}}

    However:

        {'component': {'key': {'_gpg': 'PGP Block ...', 'nothing': 'should go here'}}}

    will transform to:

        {'component': {'key': 'decrypted value'}}
    """
    def decrypt(obj):
        """Decrypt the object.

        It is an inner function because we must first verify that gpg
        is ready. If we did them in the same function we would end up
        calling the gpg checks several times, potentially, since we are
        calling this recursively.
        """
        if isinstance(obj, list):
            res_v = []
            for item in obj:
                res_v.append(decrypt(item))
            return res_v
        elif isinstance(obj, dict):
            if '_gpg' in obj:
                try:
                    decrypted = gpg.decrypt(obj['_gpg'])
                    if decrypted.ok:
                        obj = n(decrypted.data.decode('utf-8').encode())
                    else:
                        log.error("gpg error unpacking secrets %s", decrypted.stderr)
                except Exception as err:
                    log.error("error unpacking secrets %s", err)
            else:
                for k, v in obj.items():
                    obj[k] = decrypt(v)
        else:
            try:
                if 'BEGIN PGP' in obj:
                    try:
                        decrypted = gpg.decrypt(obj)
                        if decrypted.ok:
                            obj = n(decrypted.data.decode('utf-8').encode())
                        else:
                            log.error("gpg error unpacking secrets %s", decrypted.stderr)
                    except Exception as err:
                        log.error("error unpacking secrets %s", err)
            except TypeError:
                log.debug('Pass on decryption. Only decrypt strings')
        return obj

    if GPG_IMPORTED:
        if not gpg_config:
            gpg_config = {}
            defaults = {'homedir': '~/.gnupg/'}
            env_fields = {'homedir': 'FIGGYPY_GPG_HOMEDIR',
                          'binary': 'FIGGYPY_GPG_BINARY',
                          'keyring': 'FIGGYPY_GPG_KEYRING'}
            for k, v in env_fields.items():
                gpg_config[k] = env_or_default(v, defaults[k] if k in defaults else None)
        try:
            gpg = gnupg.GPG(**gpg_config)
        except (OSError, RuntimeError):
            log.exception('Failed to configure gpg. Will be unable to decrypt secrets.')
        return decrypt(cfg)
    return cfg


def kms_decrypt(cfg, aws_config=None):
    """Decrypt KMS objects in configuration.

    Args:
        cfg (dict): configuration dictionary
        aws_config (dict): aws credentials
            dict of arguments passed into boto3 session
            example:
                aws_creds = {'aws_access_key_id': aws_access_key_id,
                             'aws_secret_access_key': aws_secret_access_key,
                             'region_name': 'us-east-1'}

    Returns:
        dict: decrypted configuration dictionary

    AWS credentials follow the standard boto flow. Provided values first,
    followed by environment, and then configuration files on the machine.
    Ideally, one would set up an IAM role for this machine to authenticate.

    The aim is to find in the dictionary items which have been encrypted
    with KMS, then decrypt them if possible.

    A user can create a key "_kms" in which to store the data. All data
    at this level will be replaced with the decrypted contents. For example:

        {'component': {'key': {'_kms': 'encrypted cipher text', 'nothing': 'should go here'}}}

    will transform to:

        {'component': {'key': 'decrypted value'}}

    To get the value to be stored as a KMS encrypted string:

        from figgypy.utils import kms_encrypt
        encrypted = kms_encrypt('your secret', 'your key or alias', optional_aws_config)
    """
    def decrypt(obj):
        """Decrypt the object.

        It is an inner function because we must first configure our KMS
        client. Then we call this recursively on the object.
        """
        if isinstance(obj, list):
            res_v = []
            for item in obj:
                res_v.append(decrypt(item))
            return res_v
        elif isinstance(obj, dict):
            if '_kms' in obj:
                try:
                    res = client.decrypt(CiphertextBlob=b64decode(obj['_kms']))
                    obj = n(res['Plaintext'])
                except ClientError as err:
                    if 'AccessDeniedException' in err.args[0]:
                        log.warning('Unable to decrypt %s. Key does not exist or no access', obj['_kms'])
                    else:
                        raise
            else:
                for k, v in obj.items():
                    obj[k] = decrypt(v)
        else:
            pass
        return obj
    try:
        aws = boto3.session.Session(**aws_config)
        client = aws.client('kms')
    except NoRegionError:
        log.info('Missing or invalid aws configuration. Will not be able to unpack KMS secrets.')
        return cfg
    return decrypt(cfg)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from future.utils import bytes_to_native_str as n

from base64 import b64encode
import os

import boto3


def env_or_default(var, default=None):
    """Get environment variable or provide default.

    Args:
        var (str): environment variable to search for
        default (optional(str)): default to return
    """
    if var in os.environ:
        return os.environ[var]
    return default


def kms_encrypt(value, key, aws_config=None):
    """Encrypt and value with KMS key.

    Args:
        value (str): value to encrypt
        key (str): key id or alias
        aws_config (optional[dict]): aws credentials
            dict of arguments passed into boto3 session
            example:
                aws_creds = {'aws_access_key_id': aws_access_key_id,
                             'aws_secret_access_key': aws_secret_access_key,
                             'region_name': 'us-east-1'}

    Returns:
        str: encrypted cipher text
    """
    aws_config = aws_config or {}
    aws = boto3.session.Session(**aws_config)
    client = aws.client('kms')
    enc_res = client.encrypt(KeyId=key,
                             Plaintext=value)
    return n(b64encode(enc_res['CiphertextBlob']))



import os
import unittest

import figgypy.config


class TestConfig(unittest.TestCase):
    def test_config_pass_on_int(self):
        os.environ['FIGGYPY_GPG_HOMEDIR']='tests/resources/test-keys'
        c = figgypy.config.Config('tests/resources/test-config.yaml')
        self.assertEqual(c.number, 1)

    def test_config_load_with_gpg(self):
        os.environ['FIGGYPY_GPG_HOMEDIR'] = 'tests/resources/test-keys'
        c = figgypy.config.Config('tests/resources/test-config.yaml')
        self.assertEqual(c.db['host'], 'db.heck.ya')
        self.assertEqual(c.db['pass'], 'test password')

    def test_config_load_without_gpg(self):
        figgypy.decrypt.GPG_IMPORTED = False
        c = figgypy.config.Config('tests/resources/test-config.yaml')
        encrypted_password = (
            '-----BEGIN PGP MESSAGE-----\n'
            'Version: GnuPG v2\n'
            '\n'
            'hQIMAzf92ZrOUZL3ARAAgWexav8+pc2lnqISEuQafFZrqYI0pU3xCuMXnFZp+hpU\n'
            'gb0LsaExZ136p4ATIinFHuaLt94hFx7gULgqoSigt/2fubnUCsOGedq122xYZdtV\n'
            'Ep/24WPVQPcMVIP9pDTJTk82A41BQsOrVYorAGjjB13zFizizYHApNTcWKr4/gfR\n'
            'jmCqAX5qusXB84fXBecCJ886uEQI2v7+Vxnk+fQMqNt3ybd/uLuBLShMSygr6uLX\n'
            'zktyeZvP2QqPSWe0OpttdcvD792/SI/CTznsjbMe0wr1L81csEQcj++4o5wJop3Y\n'
            'mbQvG/FxeDdRi2aCxh7JK2xdCsrQzXKTNG2QZMwWqatB5Lb6lJ1mNiJQGX2YK+nI\n'
            'lbjy5Cp2lHlNxa9QfB+KglueMnH9gDku5YqBDos6rCEuqK/aTDdMx0V7YGYTamZ3\n'
            '3Za+OGi+hl/+4WX2gm+bOM2WWrIysiu9k1HMI1/onui/3hr1nClR8rGb4a5qDlpg\n'
            'yRrt7LuLRU4vGXpYm05dXlUeI3uT04ur/DwLo32ujnPo3dc8LFegX8N8p1LLS9vq\n'
            'vvrvXRnWsgeAvAYFBprbEYcz7sOU04HM9OGcyjYREMs3Ih6H2oBi3GavJ2x0MG75\n'
            'M9JSTu/yytD8GCM3s+3RncKuEAxfZIk1Gbdz0pjb+U6G43qq8/vQPKtKuAeqJHDS\n'
            'SAER9YkKqbp0y85LbhUWNWPpHQ2zy8WB71TfYE6vBP5qjoxiqP/QGWjT/3jhCY+t\n'
            '5k7R6XqvdvbSu1avFlEgApknzn94I+gsWQ==\n'
            '=QuDe\n'
            '-----END PGP MESSAGE-----\n'
        )
        self.assertEqual(c.db['host'], 'db.heck.ya')
        self.assertEqual(c.db['pass'].rstrip('\n'), encrypted_password.rstrip('\n'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from future.utils import bytes_to_native_str as n

from base64 import b64decode
import os

import boto3
import pytest

from figgypy.utils import kms_encrypt


@pytest.mark.skipif(os.environ.get('INTEGRATION') is None,
                    reason="credentials are required")
class TestEncryptIntegration(object):
    def test_kms_encrypt(self):
        key = 'alias/figgypy-test'
        secret = 'test password 1234567890 !@#$%^&*()'
        client = boto3.client('kms')
        encrypted = kms_encrypt(secret, key)
        dec_res = client.decrypt(CiphertextBlob=b64decode(encrypted))
        decrypted = n(dec_res['Plaintext'])
        assert decrypted == secret



import logging
import os
import yaml

import seria

from figgypy.decrypt import (
    gpg_decrypt,
    kms_decrypt
)
from figgypy.exceptions import FiggypyError

log = logging.getLogger('figgypy')
if len(log.handlers) == 0:
    log.addHandler(logging.NullHandler())


class Config(object):
    """Configuration object

    Args:
        f (str): filename
            see below
        aws_config (dict): aws credentials
            dict of arguments passed into boto3 session
            example:
                aws_creds = {'aws_access_key_id': aws_access_key_id,
                             'aws_secret_access_key': aws_secret_access_key,
                             'region_name': 'us-east-1'}
        gpg_config (dict): gpg configuration
            dict of arguments for gpg including:
                homedir, binary, and keyring (require all if any)
            example:
                gpg_config = {'homedir': '~/.gnupg/',
                              'binary': 'gpg',
                              'keyring': 'pubring.kbx'}

    Returns:
        object: configuration object with attribute dictionary for each
            top level property

    Object can be created with a filename only, relative path, or absolute path.
    If only name or relative path is provided, look in this order:

    1. current directory
    2. `~/.config/<file_name>`
    3. `/etc/<file_name>`

    It is a good idea to include you __package__ in the file name.
    For example, `cfg = Config(os.path.join(__package__, 'config.yaml'))`.
    This way it will look for your_package/config.yaml,
    ~/.config/your_package/config.yaml, and /etc/your_package/config.yaml.
    """
    _dirs = [
        os.curdir,
        os.path.join(os.path.expanduser("~"), '.config'),
        "/etc/"
    ]

    def __init__(self, f, aws_config=None, gpg_config=None):
        self._aws_config = aws_config or {}
        self._gpg_config = gpg_config or {}
        self._f = self._get_file(f)
        self.values = self._load(self._f)

    def _load(self, f):
        """Get values from config file"""
        try:
            with open(f, 'r') as _fo:
                _seria_in = seria.load(_fo)
                _y = _seria_in.dump('yaml')
        except IOError:
            raise FiggypyError("could not open configuration file")

        _values = yaml.load(_y)
        self._post_load_process(_values, self._gpg_config)
        for k, v in _values.items():
            setattr(self, k, v)
        return _values

    def _post_load_process(self, obj, gpg_config=None):
        gpg_decrypt(obj, self._gpg_config)
        kms_decrypt(obj, self._aws_config)
        return obj

    def _get_file(self, f):
        """Get a config file if possible."""
        if os.path.isabs(f):
            return f
        else:
            for d in Config._dirs:
                _f = os.path.join(d, f)
                if os.path.isfile(_f):
                    return _f
            raise FiggypyError("could not find configuration file {} in dirs {}"
                               .format(f, Config._dirs))

    def get(self, *args, **kwargs):
        """Get from values dictionary by exposing self.values.get method.

        dict.get() method on Config.values
        """
        return self.values.get(*args, **kwargs)



"""figgypy is a simple configuration manager"""

__title__ = 'figgypy'
__author__ = 'Herkermer Sherwood'

from figgypy.config import Config

_config = None


def get_config():
    """Get the global configuration.

    For this to work you must first call figgypy.set_config. See set_config for help.

    The only purpose of this helper, is so that we can raise an error telling the library
    user that they must run figgypy.set_config first. If we had them use just use
    figgypy._config, the initial value of None would give no indication of how to
    initialize the Config object.
    """
    global _config
    if _config is None:
        raise ValueError('configuration not set; run figgypy.set_config first')
    return _config


def set_config(*args, **kwargs):
    """Set a global config.

    See help(figgypy.config.Config) for full signature.

    For this to work properly you must import the full package namespace.
        # a.py
        import figgypy
        cfg = figgypy.set_cfg(cfg_file)
        # cfg is from the file you passed in

        # b.py
        import figgypy
        cfg = figgypy.get_cfg()
        # cfg is from the file you passed in still

    You will get new instances if you import from the namespace. i.e.
        # a.py
        from figgypy import set_cfg
        cfg = set_cfg(cfg_file)
        # cfg is the file you passed in

        # b.py
        from figgypy import get_cfg
        cfg = get_cfg()
        # cfg will raise an error since it has not yet been set
    """
    global _config
    _config = Config(*args, **kwargs)
    return _config


def get(*args, **kwargs):
    """Get from config object by exposing Config.values.get method.

    dict.get() method on Config.values
    """
    global _config
    if _config is None:
        raise ValueError('configuration not set; must run figgypy.set_config first')
    return _config.get(*args, **kwargs)


__all__ = ['Config', 'get', 'get_config', 'set_config']
